/*--------------------------------------------------------*/
/*   Texture Synthesis by Image Quilting                  */
/*   Date : 2004			                              */
/*--------------------------------------------------------*/
#include "bitmap.h" // for bitmap processing
#include <iostream> // for input/output stream

#include <time.h> // for random
#include <math.h> // for square calcuration

// Main Function
// To reduce the execute time, I'didn't the sub-function.
void main(int argc, char **argv)
{
	if (argc != 5) // Check the number of input arguments
	{
		std::cout << "Usage: bitmap subblock-size overlapping-size loop-count bmpfilename\n"; 
		exit(0); // quit the program
	}

	// Variables related to Bitmap Size
	const int desWidth = 384, desHeight = 384; // output bitmap image width and height size
	int subBlkWidth = atoi(argv[1]), subBlkHeight = atoi(argv[1]); // Save the input data(Sub-block Size) that received from command line(from user)
	int orgWidth, orgHeight; // input bitmap image width and height size
	long TotalPixel = desWidth * desHeight; // Output Image(Bitmap) Size = Total Pixel Number
	long subBlkTotalPixel = subBlkWidth * subBlkHeight; // Sub=block Bitmap Size
	int overlapWidth = atoi(argv[2]), overlapHeight = atoi(argv[2]); // overlaping image width & height size
	int subBlkPosX, subBlkPosY; // Save the random position(the sub-block start position) in the input image file 

	// Pointer Variables that related to bitmap
	unsigned char *desBitmap; // pointer variable for output bitmap information
	unsigned char *orgBitmap; // pointer variable for input bitmap information
	unsigned char *subBlkBmp; // pointer variable for sub-block bitmap information
	unsigned char *subBlkBmpOld, *subBlkBmpOld2, *subBlkBmpTemp; // etc. for bitmap informat

	// Dynamic Memory Allocation
	subBlkBmp = (unsigned char*)malloc(subBlkWidth * subBlkHeight * 4);     // memory allocation for bitmap data
	subBlkBmpTemp = (unsigned char*)malloc(subBlkWidth * subBlkHeight * 4); // Size : Sub-Block Width X Height
	subBlkBmpOld = (unsigned char*)malloc(subBlkWidth * subBlkHeight * 4);  // To save the RGB(And Alpha) Color,
	subBlkBmpOld2 = (unsigned char*)malloc(subBlkWidth * subBlkHeight * 4); // We have to allocate the memory 4-times more than Sub-Block memory size.
	desBitmap = (unsigned char*)malloc(desWidth * desHeight * 4); // memory allication for output bitmap data
	
	// Variables that help the program
	time_t t; // Time Structure
	srand(time(&t)); // We can get the real random number.
	int i, j, k, l, m, n; // variables for iteration structure
	int selectBlkCount = atoi(argv[3]); // The Count (to create the random sub-block)

	// The Rest (Variables)
	char ErrMsg[512]; // Variable for Error Message
	int totalDiff, totalDiff2, bestTotalDiff, bestTotalDiff2; // Get the informatio about Color Difference.

	/* Load input bitmap and The information is saved to orgBitmap. */
	/* And We get the input bitmap width & height information */
	if (!(orgBitmap = LoadBitmap(argv[4], ErrMsg, orgWidth, orgHeight))) 
	{
		std::cout << "LoadBitmap Error: " << ErrMsg << "\n"; // Fail to load the image
		exit(0); // quit the program
	}

	/* Check the overlap size */
	/* overlap region size have to be smaller than sub-block image size */
	if(subBlkWidth < overlapWidth || subBlkHeight < overlapHeight)
	{
		std::cout << "Please, input smaller overlap size than input-image size." << "\n";;
		exit(0);
	}

	/* overlap region size have to be smaller than input image size */
	if(orgWidth < overlapWidth || orgHeight < overlapHeight)
	{
		std::cout << "Please, input smaller overlap size than input-image size." << "\n";;
		exit(0);
	}

	/* overlap region size - (smaller than 100) */
	if(overlapWidth > 100 || overlapHeight > 100)
	{
		std::cout << "Please, input the correct overlap size. (range : 1 ~ 100)" << "\n";;
		exit(0);
	}

	std::cout << "Bitmap file " << argv[4] << " loaded ( size: " << orgWidth << " x " << orgHeight << " )\n"; // Show the Input Image Information
	std::cout << "Sub-Block Size : " << subBlkWidth << " x " << subBlkHeight << "\n"; // Show the Sub-Block Size Information
	std::cout << "Overlap Size : " << overlapWidth << "\n"; // Show the overlap size information (overlapWidth = overlapHeight <- same)
	std::cout << "Block Sampling Count : " << argv[3] << "\n\n"; // Creation sub-block time per sub-region. Sampling Count
	std::cout << "Now Synthesizing Images" << "\n\n"; // Show Just Message about the progress.

	// Compute negative image // not used in this program 
/*
	for (int i = 0; i < TotalPixel; i++) 
	{
		Bitmap[i*4] = 0xff - Bitmap[i*4];          // Red
		Bitmap[i*4+1] = 0xff - Bitmap[i*4+1];      // Green
		Bitmap[i*4+2] = 0xff - Bitmap[i*4+2];      // Blue
		//Bitmap[i*4+3] = Bitmap[i*4+3];             // Alpha: no change
	} 
*/

	/* Main Loop (Algorithm , I explain the detail algorithm in the report) */
	/* All the works progress in this loop. */
	for(k = 0; k <= (desHeight / (subBlkHeight - overlapHeight)); k++) // The Count that - Output Height divide by Sub-block Height, considerate the overlap size
	{
		for(i = 0; i <= (desWidth / (subBlkWidth - overlapWidth)); i++) // The Count that - Output Width divide by Sub-block Height, considerate the overlap size
		{
			// The variables that save the Best Sum of the square of color difference in the overlap region		
			// bestTotalDiff : is related the sum of left-right block
			// bestTotalDiff2 : is related the sum of top-bottom block
			bestTotalDiff = bestTotalDiff2 = 99999999; // initialize to the 99999999(it means maximum value)
			
			for(l = 0; l < selectBlkCount; l++) // Find the Random sub-block from the input image. (selectBlkCount is user-input-data. 3rd parameter)
			{
				totalDiff = totalDiff2 = 0; // initialize color difference variables

				subBlkPosX = rand() % (orgWidth - subBlkWidth); // Get the random block x-position
				subBlkPosY = rand() % (orgHeight - subBlkHeight); // Get the random block y-position

				for(j = 0; j < subBlkTotalPixel; j++) // Separate the Sub-block Image from the Input Image
				{                                     // Copy the sub-image to 'subBlkBmpTemp', Copy Direction : Left-Bottom -> Right-Top Dirction
					subBlkBmpTemp[j * 4] = orgBitmap[((((subBlkPosY) * orgWidth + subBlkPosX) + ((j / subBlkWidth) * orgWidth) + (j % subBlkWidth))) * 4]; // Red Color Copy
					subBlkBmpTemp[j * 4 + 1] = orgBitmap[((((subBlkPosY) * orgWidth + subBlkPosX) + ((j / subBlkWidth) * orgWidth) + (j % subBlkWidth))) * 4 + 1]; // Green
					subBlkBmpTemp[j * 4 + 2] = orgBitmap[((((subBlkPosY) * orgWidth + subBlkPosX) + ((j / subBlkWidth) * orgWidth) + (j % subBlkWidth))) * 4 + 2]; // Blue
				}

				/* Calculate the cost about the color difference between two sub-block in the overlapped region. */
				/* Calculate the cost between left, right block */
				for(m = 0; m < subBlkHeight; m++) // At the overlapped images
				{
					for(n = 0; n < overlapWidth; n++)
					{   // Compute the placement cost
						totalDiff += pow((subBlkBmpOld[((subBlkWidth - overlapWidth) + m * subBlkWidth + n) * 4] // get color difference count
							- subBlkBmpTemp[(m * subBlkWidth + n) * 4]), 2) // Red               
							+ pow((subBlkBmpOld[((subBlkWidth - overlapWidth) + m * subBlkWidth + n) * 4 + 1] 
							- subBlkBmpTemp[(m * subBlkWidth + n) * 4 + 1]), 2) // Green
							+ pow((subBlkBmpOld[((subBlkWidth - overlapWidth) + m * subBlkWidth + n) * 4 + 2] 
							- subBlkBmpTemp[(m * subBlkWidth + n) * 4 + 2]), 2); // Blue
					}
				}

				/* To get a overlapped sum of color difference between the top image and the bottom image, */
				/* Find the previous(top) image. We just only know the current image(bottom block) sub-block. */
				/* To get the overlapped color difference information between the top image and the bottom image, */
				/* We have to find the top image information. So I did the following work. */
				if(k != 0 && 0 < desHeight - ((subBlkHeight - overlapHeight) * k + overlapHeight)) // Check the Square Rnage // Temp : ((desHeight / (subBlkHeight - overlapHeight))) 
				{
					subBlkPosX = (subBlkWidth - overlapWidth) * i; // Get a top-image start x-position
					subBlkPosY = desHeight - ((subBlkHeight - overlapHeight) * k + overlapHeight); // Get a top-image start y-position
				}

				for(j = 0; j < subBlkTotalPixel; j++) // Using 'subBlkPosX' and 'subBlkPosY', Get a top-image RGB information, and copy it to the subBlkBmpOld2
				{
					subBlkBmpOld2[j * 4] = desBitmap[((((subBlkPosY) * desWidth + subBlkPosX) + ((j / subBlkWidth) * desHeight) + (j % subBlkWidth))) * 4];
					subBlkBmpOld2[j * 4 + 1] = desBitmap[((((subBlkPosY) * desWidth + subBlkPosX) + ((j / subBlkWidth) * desHeight) + (j % subBlkWidth))) * 4 + 1];
					subBlkBmpOld2[j * 4 + 2] = desBitmap[((((subBlkPosY) * desWidth + subBlkPosX) + ((j / subBlkWidth) * desHeight) + (j % subBlkWidth))) * 4 + 2];
				}

				if(k != 0) // if 'k' is not zero..   (if k is zero, we don't need to do this work.)
				{ 
					/* Calculate the cost about the color difference between two sub-block in the overlapped region. */
					/* Calculate the cost between top, bottom block */
					/* To this work, we already get the top-image(block) information */
					for(n = 0; n < overlapHeight; n++)
					{
						for(m = 0; m < subBlkWidth; m++)
						{   // Compute the placement cost
							totalDiff2 += pow((subBlkBmpOld2[(m + n * subBlkWidth) * 4]
								- subBlkBmpTemp[((subBlkHeight - overlapWidth) * subBlkWidth + n * subBlkWidth + m) * 4]), 2)
								+ pow((subBlkBmpOld2[(m + n * subBlkHeight) * 4 + 1]
								- subBlkBmpTemp[((subBlkHeight - overlapWidth) * subBlkWidth + n * subBlkWidth + m) * 4 + 1]), 2)
								+ pow((subBlkBmpOld2[(m + n * subBlkHeight) * 4 + 2]
								- subBlkBmpTemp[((subBlkHeight - overlapWidth) * subBlkWidth + n * subBlkWidth + m) * 4 + 2]), 2);
						}
					}
				}

				/* Find the the Optimal Cost */
				/* Considerate the Left-Right Overlapped Area's Color Difference and Top-Bottom Overlapped Area's Color Difference */
				if((bestTotalDiff + bestTotalDiff2) > (totalDiff + totalDiff2))
				{
					memcpy(subBlkBmp, subBlkBmpTemp, subBlkWidth * subBlkHeight * 4); // The Best Block is copied to 'subBlkBmp'
					bestTotalDiff = totalDiff; // Change the variables that have a optimal Cost.
					bestTotalDiff2 = totalDiff2;
				}
			}	

			// Additional Variables
			unsigned int diffValue[100][10] = {0, }; // Double-Array to save the color difference in the overlap area.
			int cutPath[100] = {0, }, cutPath2[100] = {0, }; // Array for minimum cost cut path.
			int minCost = 99999999, minPos = 0;
			int minValue = 0;

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/////////////* The Beginning of Image Quilting Work *////////////////////////////////////////////////////////////////

/////////////* Image Quilting Work 1 *///////////////////////////////////////////////////////////////////////////////
			/* Minimum Error Boundary Cut */
			/* Work for Left-Right(Horizontal) Overlapped Region */
			/* Reference : Graphcut Textures : Image and Video Synthesis Using Graph Cuts - 3. Patch Fitting using Graph Cuts */
			if(i != 0) // don't need to work about the first-horizontal line block
			{
				for(m = 0; m < subBlkHeight; m++)
				{
					diffValue[m][0] = 99999999; // The Boundary Area : '99999999' means infinite value, We use it to get a cut path weight.
					diffValue[m][overlapWidth + 1] = 99999999; // The Boundary Area
				}

				/* Get The Color Difference per pixel. We use it to find a optimal cut path in the overlapped region. */
				for(m = (subBlkHeight - 1); m >= 0; m--) // At the overlaped region
				{
					for(n = 1; n < (overlapWidth + 1); n++)
					{   // Get a color difference per pixel. And Save the data to the diffValue array.
						diffValue[m][n] += pow((subBlkBmpOld[((subBlkWidth - overlapWidth) + m * subBlkWidth + n) * 4] 
							- subBlkBmp[(m * subBlkWidth + n) * 4]), 2) 
							+ pow((subBlkBmpOld[((subBlkWidth - overlapWidth) + m * subBlkWidth + n) * 4 + 1] 
							- subBlkBmp[(m * subBlkWidth + n) * 4 + 1]), 2)
							+ pow((subBlkBmpOld[((subBlkWidth - overlapWidth) + m * subBlkWidth + n) * 4 + 2] 
							- subBlkBmp[(m * subBlkWidth + n) * 4 + 2]), 2);
					}
				}

				/* Compute the weight per pixel. I explain the detail algorithm in the report. */
				for(m = subBlkHeight - 1; m > 0; m--) 
				{
					for(n = 1; n < (overlapWidth + 1); n++)
					{
						// among high ranked adjacenct 3-pixels, find the pixel that have a minimum weight.
						if(diffValue[m][n - 1] > diffValue[m][n])
							minValue = (diffValue[m][n] > diffValue[m][n + 1] ? diffValue[m][n + 1] : diffValue[m][n]);
						else if(diffValue[m][n] > diffValue[m][n + 1])
							minValue = (diffValue[m][n + 1] > diffValue[m][n - 1] ? diffValue[m][n - 1] : diffValue[m][n + 1]);
						else if(diffValue[m][n + 1] > diffValue[m][n - 1])
							minValue = (diffValue[m][n - 1] > diffValue[m][n] ? diffValue[m][n] : diffValue[m][n - 1]);

						diffValue[m - 1][n] += minValue;
					}
				}

				/* Find the first position in the cut-path. The first position must have a minimun weight among the pixels that is positioned in a low, horizontally. */
				for(n = 1; n < (overlapWidth + 1); n++)
				{
					if(minCost > diffValue[0][n])
					{
						minCost = diffValue[0][n]; // Get a pixel color value.
						minPos = n; // Get a pixel position.
					}
				}

				cutPath[0] = minPos; // Save The first position value to the first position in the cutPath array.

				/* Set the rest pixels for minimum-cut-path. */
				/* The minimum-cut-path must connected, certainly. */
				for(m = 1; m < subBlkHeight; m++)
				{
					// among high ranked adjacenct 3-pixels, find the pixel that have a minimum weight.
					if(diffValue[m][minPos - 1] > diffValue[m][minPos])
						minPos = (diffValue[m][minPos] > diffValue[m][minPos + 1] ? minPos + 1 : minPos);
					else if(diffValue[m][minPos] > diffValue[m][minPos + 1])
						minPos = (diffValue[m][minPos + 1] > diffValue[m][minPos - 1] ? minPos - 1 : minPos + 1);
					else if(diffValue[m][minPos + 1] > diffValue[m][minPos - 1])
						minPos = (diffValue[m][minPos - 1] > diffValue[m][minPos] ? minPos : minPos - 1);

					cutPath[m] = minPos; // and add the minimum weight(value). Namely, Make the cut-line.
				}
			}
/////////////* End of First Image Quilting Work *////////////////////////////////////////////////////////////////////

/////////////* Image Quilting Work 2 *///////////////////////////////////////////////////////////////////////////////
			/* Work for Left-Right(Horizontal) Overlapped Region */
			/* It is similar 'Image Quilting Work 1' */
			if(k != 0) // don't need to work about the first-vertical line block. becaust they don't have the block to overlap.
			{
				for(m = 0; m < subBlkWidth; m++)
				{
					diffValue[m][0] = 99999999; // The Boundary Area
					diffValue[m][overlapHeight + 1] = 99999999; // The Boundary Area
				}

				/* Get The Color Difference per pixel. We use it to find a optimal cut path in the overlapped region. */
				for(m = 0; m <= (subBlkWidth - 1); m++)
				{
					for(n = 1; n < (overlapHeight + 1); n++)
					{
						// Get a color difference per pixel. And Save the data to the diffValue array.
						diffValue[m][n] += pow(subBlkBmpOld2[(m + (overlapHeight - n) * subBlkWidth) * 4] 
							- subBlkBmp[((subBlkHeight - n) * subBlkWidth + m) * 4], 2)
							+ pow(subBlkBmpOld2[(m + (overlapHeight - n) * subBlkWidth) * 4 + 1]
							- subBlkBmp[((subBlkHeight - n) * subBlkWidth + m) * 4 + 1], 2)						
							+ pow(subBlkBmpOld2[(m + (overlapHeight - n) * subBlkWidth) * 4 + 2] 
							- subBlkBmp[((subBlkHeight - n) * subBlkWidth + m) * 4 + 2], 2);
					}
				}

				/* Compute the weight per pixel. */
				for(m = 0; m < (subBlkWidth - 1); m++) 
				{
					for(n = 1; n < (overlapHeight + 1); n++)
					{
						// among high ranked adjacenct 3-pixels, find the pixel that have a minimum weight.
						if(diffValue[m][n - 1] > diffValue[m][n])
							minValue = (diffValue[m][n] > diffValue[m][n + 1] ? diffValue[m][n + 1] : diffValue[m][n]);
						else if(diffValue[m][n] > diffValue[m][n + 1])
							minValue = (diffValue[m][n + 1] > diffValue[m][n - 1] ? diffValue[m][n - 1] : diffValue[m][n + 1]);
						else if(diffValue[m][n + 1] > diffValue[m][n - 1])
							minValue = (diffValue[m][n - 1] > diffValue[m][n] ? diffValue[m][n] : diffValue[m][n - 1]);
	
						diffValue[m + 1][n] += minValue;
					}
				}

				minCost = 99999999, minPos = 0; // Initialize the minCost and minPos

				/* Find the first position in the cut-path. The first position must have a minimun weight among the pixels that is positioned in a low, horizontally. */
				for(n = 1; n < (overlapHeight + 1); n++)
				{
					if(minCost > diffValue[subBlkWidth - 1][n])
					{
						minCost = diffValue[subBlkWidth - 1][n];
						minPos = n;
					}
				}

				cutPath2[47] = minPos; // Save The first position value to the last position in the cutPath array.

				/* Set the rest pixels for minimum-cut-path. */
				/* The minimum-cut-path must connected, certainly. */
				for(m = (subBlkWidth - 1) - 1; m >= 0; m--)
				{
					// among high ranked adjacenct 3-pixels, find the pixel that have a minimum weight.
					if(diffValue[m][minPos - 1] > diffValue[m][minPos])
						minPos = (diffValue[m][minPos] > diffValue[m][minPos + 1] ? minPos + 1 : minPos);
					else if(diffValue[m][minPos] > diffValue[m][minPos + 1])
						minPos = (diffValue[m][minPos + 1] > diffValue[m][minPos - 1] ? minPos - 1 : minPos + 1);
					else if(diffValue[m][minPos + 1] > diffValue[m][minPos - 1])
						minPos = (diffValue[m][minPos - 1] > diffValue[m][minPos] ? minPos : minPos - 1);
	
					cutPath2[m] = minPos; // and add the minimum weight(value). Namely, Make the cut-line.
				}
			}
/////////////* End of Second Image Quilting Work *///////////////////////////////////////////////////////////////////

/////////////* The Final Work for Image Quilting *///////////////////////////////////////////////////////////////////
			/* Using the minimum-cut-path information, split the overlapped region to the two area(A - B) */ 
			if(i != 0) // Except the first-line blocks (horizontal) 
			{
				for(m = 0; m < subBlkHeight; m++)
				{
					for(n = 0; n <= cutPath[m]; n++) //  о  ߶󳻱
					{
						subBlkBmp[(m * subBlkHeight + n) * 4] = subBlkBmpOld[((subBlkWidth - overlapWidth) + m * subBlkHeight + n) * 4];
						subBlkBmp[(m * subBlkHeight + n) * 4 + 1] = subBlkBmpOld[((subBlkWidth - overlapWidth) + m * subBlkHeight + n) * 4 + 1];
						subBlkBmp[(m * subBlkHeight + n) * 4 + 2] = subBlkBmpOld[((subBlkWidth - overlapWidth) + m * subBlkHeight + n) * 4 + 2];
					}
				}
			}

			/* Using the minimum-cut-path information, split the overlapped region to the two area(A - B) */
			if(k != 0) // Escept the first-line blockx (vertical)
			{
				for(m = 0; m < subBlkWidth; m++)
				{
					for(n = 0; n < cutPath2[m]; n++)
					{
						subBlkBmp[((subBlkHeight - (n + 1)) * subBlkWidth + m) * 4] = subBlkBmpOld2[((overlapHeight - (n + 1)) * subBlkWidth + m) * 4];
						subBlkBmp[((subBlkHeight - (n + 1)) * subBlkWidth + m) * 4 + 1] = subBlkBmpOld2[((overlapHeight - (n + 1)) * subBlkWidth + m) * 4 + 1];
						subBlkBmp[((subBlkHeight - (n + 1)) * subBlkWidth + m) * 4 + 2] = subBlkBmpOld2[((overlapHeight - (n + 1)) * subBlkWidth + m) * 4 + 2];
					}
				}
			}
/////////////* The End of All Image Quilting Work *//////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

			if(i == 0) 	overlapWidth = 0; // In case of the block that is positioned first-line (horizontal), don't need to use overlapWidth
			else		overlapWidth = atoi(argv[2]);

			if(k == 0)  overlapHeight = 0; 
			else        overlapHeight = atoi(argv[2]);

			/* Paste the Sub-blocks to the output-image(desBitmap) */
			for(j = 0; j < subBlkTotalPixel; j++)
			{
				if((subBlkWidth - overlapWidth) * i + (j % subBlkWidth) >= desWidth) // if the position excess the width-length
					continue;

				if(desHeight - (subBlkHeight * (1 + k)) + (overlapHeight * k) + (j / subBlkHeight) < 0) // if the position excess the height-length
					continue;

				/* Copy the sub-block to the output-image */
				desBitmap[((((desHeight - (subBlkHeight * (1 + k)) + (overlapHeight * k)) * desWidth + ((subBlkWidth - overlapWidth) * i)) + (j / subBlkHeight) * desWidth + (j % subBlkWidth))) * 4] = subBlkBmp[j * 4];          // Red
				desBitmap[((((desHeight - (subBlkHeight * (1 + k)) + (overlapHeight * k)) * desWidth + ((subBlkWidth - overlapWidth) * i)) + (j / subBlkHeight) * desWidth + (j % subBlkWidth))) * 4 + 1] = subBlkBmp[j * 4 + 1];      // Green
				desBitmap[((((desHeight - (subBlkHeight * (1 + k)) + (overlapHeight * k)) * desWidth + ((subBlkWidth - overlapWidth) * i)) + (j / subBlkHeight) * desWidth + (j % subBlkWidth))) * 4 + 2] = subBlkBmp[j * 4 + 2];      // Blue
				//desBitmap[i*4+3] = Bitmap[i*4+3];             // Alpha: no change
			}

			memcpy(subBlkBmpOld, subBlkBmp, subBlkWidth * subBlkHeight * 4); // Keep the previous sub-block, To use it for the overlap-region comparison
			overlapWidth = overlapHeight = atoi(argv[2]); // If the overlap value is zero, Restore it.
		}
	}

	WriteBitmapFile("output.bmp", desWidth, desHeight, desBitmap, 32, ErrMsg); // create the output-image file

	std::cout << "Work is completed." << "\n"; // Message to inform the end of program

	delete[] orgBitmap; // Memory Deallocation (input bitmap)
	delete[] subBlkBmp; // (Sub-Block bitmap)
	delete[] desBitmap; // (output bitmap)
}